#include <grp.h>
#include <sys/types.h>
#include <sys/file.h>
#include "untamo.h"
#include "y.tab.h"

#define EXMPT 'E'
#define RULE 'R'

extern char *strcpy(), *ctime();
extern strlen();
time_t time();
long lseek();

/*
 * addlist -- adds a record to the list "list".  Uses the 
 *	      insque(3) call for speed.
 *	      list -- which list to add the rule to (rules or exempt)
 *	      type -- what kind of rule (LOGIN, GROUP, etc)
 *	      name -- who it applies to (ajw, console, etc)
 *	      num --  who it applies to (1505 (rld), 5 (group staff) )
 *	      flag -- idle time for a "rule" rule, or exemption type
 *		      for "exempt" rule: IDLE, MULTIPLE, etc.
 */
addlist(list, type, name, num, flag)
struct qelem *list;
char *name;
int type, flag, num;
{
	struct qelem *new_node;
	struct qelem *ptr;
	struct item  *new_data;

	/* 
	 * make all the new structures
	 */
	new_node = (struct qelem *) malloc( sizeof(struct qelem) );
	new_data = (struct item  *) malloc( sizeof(struct item ) );
	new_data->name_t = type;
	new_data->name = name;
	new_data->num = num;
	new_data->flag = flag;
	new_node->q_item = new_data; 
	/*
	 * find where to insert it in the list
	 */
	ptr = list->q_forw;
	while (ptr != list)	{
		if (ptr->q_item->name_t <= new_data->name_t)
			break;
		ptr = ptr->q_forw;
	}
	/*
	 * and insert it
	 */
	(void) insque(new_node,ptr->q_back);
}


/*
 * freelist -- frees up the space in the list pointed to 
 *	       by ptr.  Uses free.
 */
freelist(ptr)
struct qelem *ptr;
{
	struct qelem *dead;
	struct qelem *start;

	start = ptr;
	ptr = ptr->q_forw;
	while ( ptr != start )   {
		dead = ptr;
		ptr = ptr->q_forw;
		(void) free( (char *) dead->q_item);  	/* kill the data */
		(void) free( (char *) dead );		/* now get the node */
	}
	start->q_forw = start;		/* reset pointers for a null list */
	start->q_back = start;
}


/*
 * setlimits -- looks through the rules list and uses the most
 *		specific rule for users[i], then looks through
 *		the exceptions and uses the (again) most specific
 *		exemption for the user.
 */

setlimits(i)
int i;
{
	struct qelem *ptr;
	int rule;
	time_t tempus;

	/*
	 * look down the rules list and set the
	 * most specific rule that applies to 
	 * this user
	 */
	rule = 0;
	(void) time(&tempus);
	ptr = rules->q_back;		/* start at the end of the list */
	users[i].warned = 0;		/* clear his warning and exempt flag */
	users[i].wcount = 0;
	users[i].exempt = 0;		/* to avoid granularity problems */
	users[i].session = 0;
	users[i].idle = 0;

	users[i].next = tempus;		/* next time is set to now, so */
					/* the new rules (or exemptions) */
					/* take affect immediately */
	/* 
	 * while we haven't looked through the whole list,
	 * and we haven't found a rule...
	 */
	while ( (ptr != rules) && (!rule) )	{
		if ( (rule = find(ptr, i, RULE)) )  {
		    if (users[i].idle < (ptr->q_item)->flag * 60) 
			users[i].idle = (ptr->q_item)->flag * 60;
			users[i].exempt &= ~IS_IDLE;
		} else
			ptr = ptr->q_back;		/* move back one */
	}

	rule = 0;
	ptr = session->q_back;
	while ( (ptr != session) && (!rule) ){
		if ( (rule = find(ptr, i, RULE)) )  {
			users[i].session = (ptr->q_item)->flag * 60;
			users[i].exempt &= ~IS_LIMIT;
		} else
			ptr = ptr->q_back;		/* move back one */
	}
	/*
	 * now look down the exemptions list and
	 * set the first exemptions for this user.
	 */
	rule = 0;
	ptr = exmpt->q_back;		/* start at the end of the list */
	while ( (ptr != exmpt) && (!rule)) {
		if ( (rule = find(ptr, i, EXMPT)) )  {
			switch ((ptr->q_item)->flag) {
				case ALL:
					users[i].exempt |=
					    IS_IDLE|IS_MULT|IS_LIMIT;
					break;
				case IDLE:
					users[i].exempt |= IS_IDLE;
					break;
				case MULTIPLE:
					users[i].exempt |= IS_MULT;
					break;
				case SESSION:
					users[i].exempt |= IS_LIMIT;
					break;
		        }
		} else
			ptr = ptr->q_back;		/* move back one */
	}
}


/*
 * find -- given a rule (or exemption) and a users, see if it applies.
 *	   Example: If the rule is a LOGIN type rule, compare the name with that
 *	   of the user to see if it applies.  If we find a match, return 1,
 *	   else return 0.
 */
find(ptr,i, kind)
int i;
struct qelem *ptr;
int kind;

{
    int x;

	switch ( (ptr->q_item)->name_t )	{
	case LOGIN: 
		if (!strcmp( (ptr->q_item)->name, users[i].uid) )
			return(1);
		break;
	case TTY:
		if (!strcmp( (ptr->q_item)->name, (users[i].line+5) ) ) 
			return(1);
		break;
	case GROUP:
		for (x=0; x<users[i].ngrps; x++) {
		    if (users[i].ugroup[x] == (ptr->q_item)->num)
			if (users[i].ngrps == x) {
			    return 1;
			} else {
			    if (kind == EXMPT) {
				switch ((ptr->q_item)->flag) {
				  case ALL:
				    users[i].exempt |= IS_IDLE|IS_MULT|IS_LIMIT;
				    break;
				  case IDLE:
				    users[i].exempt |= IS_IDLE;
				    break;
				  case MULTIPLE:
				    users[i].exempt |= IS_MULT;
				    break;
				  case SESSION:
				    users[i].exempt |= IS_LIMIT;
				    break;
				}
				return 0;
			    } else {
				if (users[i].idle < (ptr->q_item)->flag * 60)
				    users[i].idle = (ptr->q_item)->flag * 60;
			    }
			}
		}
		break;
	case DEFAULT:
		return(1);
		break;
	}
	return(0);
}

